#include "Renderer.h"
#include "GeometryNode.h"
#include "Tools.h"
#include "ShaderProgram.h"
#include "glm/gtc/type_ptr.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "OBJLoader.h"

#include <algorithm>
#include <array>

// RENDERER
Renderer::Renderer()
{
	this->m_nodes = {};
	this->m_continous_time = 0.0;
}

Renderer::~Renderer()
{
	// empty
}

bool Renderer::Init(int SCREEN_WIDTH, int SCREEN_HEIGHT)
{
	this->m_screen_width = SCREEN_WIDTH;
	this->m_screen_height = SCREEN_HEIGHT;

	bool techniques_initialization = InitShaders();
	bool meshes_initialization = InitGeometricMeshes();

	//If there was any errors
	if (Tools::CheckGLError() != GL_NO_ERROR)
	{
		printf("Exiting with error at Renderer::Init\n");
		return false;
	}

	if (meshes_initialization)
	{
		this->BuildWorld();
		this->InitCamera();
	}

	//If everything initialized
	return techniques_initialization && meshes_initialization;
}

void Renderer::BuildWorld()
{
	GeometryNode& bunny = *this->m_nodes[OBJECS::BUNNY];
	GeometryNode& ball = *this->m_nodes[OBJECS::BALL];
	GeometryNode& chair = *this->m_nodes[OBJECS::CHAIR];
	GeometryNode& floor = *this->m_nodes[OBJECS::FLOOR];

	ball.model_matrix = glm::translate(glm::mat4(1.f), glm::vec3(0.f, 35.f, 0.f));
	ball.m_aabb.center = glm::vec3(ball.model_matrix * glm::vec4(ball.m_aabb.center, 1.f));

	glm::mat4 R = glm::rotate(glm::mat4(1.f), glm::radians(-90.f), glm::vec3(0.f, 1.f, 0.f));

	bunny.model_matrix =
		glm::translate(glm::mat4(1.f), glm::vec3(bunny.m_aabb.center.x, bunny.m_aabb.center.y, bunny.m_aabb.center.z)) *
		R * 
		glm::translate(glm::mat4(1.f), glm::vec3(-bunny.m_aabb.center.x, -bunny.m_aabb.center.y, -bunny.m_aabb.center.z));

	chair.model_matrix = glm::mat4(1.f);
	floor.model_matrix = glm::mat4(1.f);

	this->m_world_matrix = glm::scale(glm::mat4(1.f), glm::vec3(0.02, 0.02, 0.02));
}

void Renderer::InitCamera()
{
	this->m_camera_position = glm::vec3(0, 3, 6);
	this->m_camera_target_position = glm::vec3(0, 0, 0);
	this->m_camera_up_vector = glm::vec3(0, 1, 0);

	this->m_view_matrix = glm::lookAt(
		this->m_camera_position,
		this->m_camera_target_position,
		m_camera_up_vector);

	this->m_projection_matrix = glm::perspective(
		glm::radians(45.f),
		this->m_screen_width / (float) this->m_screen_height,
		0.1f, 1000.f);
}

bool Renderer::InitShaders()
{
	std::string vertex_shader_path = "Assets/Shaders/basic_rendering.vert";
	std::string fragment_shader_path = "Assets/Shaders/basic_rendering.frag";

	this->m_geometry_rendering_program.LoadVertexShaderFromFile(vertex_shader_path.c_str());
	this->m_geometry_rendering_program.LoadFragmentShaderFromFile(fragment_shader_path.c_str());
	this->m_geometry_rendering_program.CreateProgram();
	this->m_geometry_rendering_program.LoadUniform("uniform_projection_matrix");
	this->m_geometry_rendering_program.LoadUniform("uniform_normal_matrix");

	return true;
}

bool Renderer::InitGeometricMeshes()
{
	std::array<const char*, OBJECS::SIZE_ALL> assets = {
		"Assets/bunny_scene/bunny.obj",
		"Assets/bunny_scene/ball.obj",
		"Assets/bunny_scene/chair.obj",
	    "Assets/bunny_scene/floor.obj" };

	bool initialized = true;
	OBJLoader loader;

	for (auto & asset : assets)
	{
		GeometricMesh* mesh = loader.load(asset);

		if (mesh != nullptr)
		{
			GeometryNode* node = new GeometryNode();
			node->Init(mesh);
			this->m_nodes.push_back(node);
			delete mesh;
		}
		else
		{
			initialized = false;
		}
	}

	return initialized;
}

void Renderer::Update(float dt)
{
	this->UpdateGeometry(dt);
	this->UpdateCamera(dt);
	m_continous_time += dt;
}

void Renderer::UpdateGeometry(float dt)
{
	GeometryNode& bunny = *this->m_nodes[OBJECS::BUNNY];
	GeometryNode& ball = *this->m_nodes[OBJECS::BALL];
	GeometryNode& chair = *this->m_nodes[OBJECS::CHAIR];
	GeometryNode& floor = *this->m_nodes[OBJECS::FLOOR];

	bunny.app_model_matrix =
		glm::translate(glm::mat4(1.f), glm::vec3(bunny.m_aabb.center.x, bunny.m_aabb.center.y, bunny.m_aabb.center.z)) *
		glm::rotate(glm::mat4(1.f), m_continous_time, glm::vec3(0.f, 1.f, 0.f)) *
		glm::translate(glm::mat4(1.f), glm::vec3(-bunny.m_aabb.center.x, -bunny.m_aabb.center.y, -bunny.m_aabb.center.z)) *
		bunny.model_matrix;

	ball.app_model_matrix =
		glm::translate(glm::mat4(1.f), glm::vec3(ball.m_aabb.center.x, ball.m_aabb.center.y, ball.m_aabb.center.z)) *
		glm::rotate(glm::mat4(1.f), m_continous_time, glm::vec3(.5f, .5f, 0.f)) *
		glm::translate(glm::mat4(1.f), glm::vec3(-ball.m_aabb.center.x, -ball.m_aabb.center.y, -ball.m_aabb.center.z)) *
		ball.model_matrix;
}

void Renderer::UpdateCamera(float dt)
{
	glm::vec3 direction = glm::normalize(m_camera_target_position - m_camera_position);

	m_camera_position = m_camera_position + (m_camera_movement.x * 5.f * dt) * direction;
	m_camera_target_position = m_camera_target_position + (m_camera_movement.x * 5.f * dt) * direction;

	glm::vec3 right = glm::normalize(glm::cross(direction, m_camera_up_vector));

	m_camera_position = m_camera_position + (m_camera_movement.y * 5.f * dt) * right;
	m_camera_target_position = m_camera_target_position + (m_camera_movement.y * 5.f * dt) * right;

	float speed = glm::pi<float>() * 0.0002;
	glm::mat4 rotation = glm::rotate(glm::mat4(1.f), m_camera_look_angle_destination.y * speed, right);
	rotation *= glm::rotate(glm::mat4(1.f), m_camera_look_angle_destination.x * speed, m_camera_up_vector);
	m_camera_look_angle_destination = glm::vec2(0.f);

	direction = rotation * glm::vec4(direction, 0.f);
	m_camera_target_position = m_camera_position + direction * glm::distance(m_camera_position, m_camera_target_position);

	m_view_matrix = glm::lookAt(m_camera_position, m_camera_target_position, m_camera_up_vector);
}

void Renderer::Render()
{
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	glClearColor(1.f, 1.f, 1.f, 1.f);
	glClear(GL_COLOR_BUFFER_BIT);

	glViewport(0, 0, m_screen_width, m_screen_height);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glClearDepth(1.f);
	glClear(GL_DEPTH_BUFFER_BIT);

	RenderGeometry();

	glDisable(GL_DEPTH_TEST);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	GLenum error = Tools::CheckGLError();
	if (error != GL_NO_ERROR)
	{
		printf("Reanderer:Draw GL Error\n");
		system("pause");
	}
}

void Renderer::RenderGeometry()
{
	this->m_geometry_rendering_program.Bind();

	glm::mat4 proj = m_projection_matrix * m_view_matrix * m_world_matrix;

	for (auto& node : this->m_nodes)
	{
		glBindVertexArray(node->m_vao);

		glUniformMatrix4fv(m_geometry_rendering_program["uniform_projection_matrix"], 1, GL_FALSE,
			glm::value_ptr(proj * node->app_model_matrix));

		glm::mat4 normal_matrix = glm::transpose(glm::inverse(m_world_matrix * node->app_model_matrix));

		glUniformMatrix4fv(m_geometry_rendering_program["uniform_normal_matrix"], 1, GL_FALSE,
			glm::value_ptr(normal_matrix));

		for (int j = 0; j < node->parts.size(); ++j)
		{
			glDrawArrays(GL_TRIANGLES, node->parts[j].start_offset, node->parts[j].count);
		}

		glBindVertexArray(0);
	}


	this->m_geometry_rendering_program.Unbind();
}

void Renderer::CameraMoveForward(bool enable)
{
	m_camera_movement.x = (enable) ? 1 : 0;
}
void Renderer::CameraMoveBackWard(bool enable)
{
	m_camera_movement.x = (enable) ? -1 : 0;
}

void Renderer::CameraMoveLeft(bool enable)
{
	m_camera_movement.y = (enable) ? -1 : 0;
}
void Renderer::CameraMoveRight(bool enable)
{
	m_camera_movement.y = (enable) ? 1 : 0;
}

void Renderer::CameraLook(glm::vec2 lookDir)
{
	m_camera_look_angle_destination = lookDir;
}